#===================================
#		   preferences
#  Copyright 2011 - Nathan Osman
#
#  A web server that allows prefs.
#   to be changed from within the
#         user's browser.
#
#   StackApplet is released under
#		the MIT license
#===================================

#TODO: add translations to this module

import BaseHTTPServer
import os
import platform
import sys
import threading
import urlparse

import gobject
gobject.threads_init()

# Now lets see if we can import the JSON
# module. We try either of these two classes
try:
	import json
except ImportError:
	import simplejson as json

# This is the callback that the main thread would like
# us to use whenever we need to issue a request for data
# from the main thread
main_thread_callback = None

def register_callback(callback):
	
	global main_thread_callback
	main_thread_callback = callback

# Here is a list of files that we can
# serve from the web server
file_list = ['/index.html',
             '/bg.png',
             '/main_logo.png',
             '/dark_theme.png',
             '/light_theme.png']

# Here are a list of information retrieving
# paths that we can respond to
data_list = ['/get_settings',
             '/add_account',
             '/remove_account',
             '/reorder',
             '/set_theme' ]

class preference_server(BaseHTTPServer.BaseHTTPRequestHandler):
	
	# There are basically two types of requests
	# that we process - GET requests (for files)
	# and POST requests (for actions).
	def do_GET(self):
		
		global file_list, data_list
		
		# self.path contains the path that we want
		# to deal with. Check to see if it's one of the
		# pages that we serve static content from.
		if self.path == '/':
			self.path = '/index.html'
		
		if self.path in file_list:
			
			self.process_file()
			
		else:
			
			self.send_reply(404, "The page you have requested does not exist on this server.")
	
	def do_POST(self):
		
		# Get the POST data
		length = int(self.headers.getheader('content-length'))
		self.post_vars = urlparse.parse_qs(self.rfile.read(length))
		
		if self.path in data_list:
			
			self.process_data()
		
		else:
			
			self.send_reply(404, "The specified command is not recognized by StackApplet.")
	
	def send_reply(self, status, reply, content_type = 'text/plain'):
		
		self.send_response(status)
		self.send_header('Content-length', str(len(reply)))
		self.send_header('Content-type', content_type)
		self.end_headers()
		self.wfile.write(reply)
	
	def process_file(self):
		
		# Now we need to simply construct the path
		# to the file and grab the contents. First check
		# to see if the file is in the html folder.
		filename = os.path.join(sys.path[0], "html" + self.path)
		
		if not os.path.isfile(filename):
			filename = os.path.join(sys.path[0], "../html" + self.path)
		
		f = open(filename, 'r')
		data = f.read()
		
		self.send_reply(200, data, 'text/html')
	
	def process_data(self):
		
		# Now we check to see which request it was
		if self.path == '/get_settings':
			self.process_get_settings()
		elif self.path == '/add_account':
			self.process_add_account()
		elif self.path == '/remove_account':
			self.process_remove_account()
		elif self.path == '/reorder':
			self.process_reorder()
		elif self.path == '/set_theme':
			self.process_set_theme()
	
	def process_get_settings(self):
		
		# We need to get the current list of accounts.
		# This means sending a request to the main thread
		# and waiting for a reply with the list of accounts.
		global main_thread_callback
		
		# Now have this message processed in the main
		# thread BUT then wait for it to be signaled
		# once processing is complete
		event = threading.Event()
		data = {}
		json_account = []
		
		gobject.idle_add(main_thread_callback, 'get_settings', event, data)
		event.wait()
		
		# Account list should now have our data
		for account in data['accounts']:
			
			json_account.append({ 'site_name': account['site_name'] })
		
		response = json.dumps({ 'accounts'    : json_account,
		                        'theme'       : data['theme'],
		                        'refresh_rate': data['refresh_rate'] })
		
		self.send_reply(200, response, 'application/json')
	
	def process_add_account(self):
		
		global main_thread_callback
		
		# Get the variables for this account
		site      = self.post_vars['site'][0]
		site_name = self.post_vars['site_name'][0]
		user_id   = self.post_vars['user_id'][0]
		
		event = threading.Event()
		gobject.idle_add(main_thread_callback, 'add_account', event, { 'site'     : site,
		                                                               'site_name': site_name,
		                                                               'user_id'  : user_id })
		event.wait()
		
		self.send_reply(200, '', 'application/json')
	
	def process_remove_account(self):
		
		global main_thread_callback
		
		account_index = self.post_vars['index'][0]
		
		event = threading.Event()
		gobject.idle_add(main_thread_callback, 'remove_account', event, account_index)
		event.wait()
		
		self.send_reply(200, '', 'application/json')
	
	def process_reorder(self):
		
		global main_thread_callback
		
		old_index = self.post_vars['old_index'][0]
		new_index = self.post_vars['new_index'][0]
		
		event = threading.Event()
		gobject.idle_add(main_thread_callback, 'reorder', event, { 'old_index': old_index,
		                                                           'new_index': new_index })
		event.wait()
		
		self.send_reply(200, '', 'application/json')
	
	def process_set_theme(self):
		
		global main_thread_callback
		
		theme = self.post_vars['theme'][0]
		
		event = threading.Event()
		gobject.idle_add(main_thread_callback, 'set_theme', event, theme)
		event.wait()
		
		self.send_reply(200, '', 'application/json')
	
	# This is here to supplant those annoying
	# status messages that display whenever a client
	# views a page.
	def log_message(self, a, *b):
		
		pass

# Create the server
httpd = BaseHTTPServer.HTTPServer(('localhost', 8140), preference_server)

# No create a special thread just for the server
thread = threading.Thread(target=httpd.serve_forever)

# Set it as a daemon and start it
thread.setDaemon(True)
thread.start()
