#===================================
#		   stackapplet
#  Copyright 2011 - Nathan Osman
#
#	 Contains the main class
#	  used by StackApplet
#
#   StackApplet is released under
#		the MIT license
#===================================

# Standard modules
import os
import platform
import signal
import sys
import time
import urllib2
import webbrowser

# The following modules are required
# for supporting translations
import locale
import gettext

# Bind the domain
gettext.bindtextdomain("stackapplet","/usr/share/locale")
gettext.textdomain("stackapplet")

# Figure out the system locale
(lang, enc) = locale.getdefaultlocale()

# Set it to a sensible default if we couldn't
# figure it out otherwise
if lang == None:
	lang = "en_US"

# Now create the translation object and
# register the translation function
tr = gettext.translation("stackapplet",
						 "/usr/share/locale",
						 [lang],
						 fallback=True)
_ = tr.gettext

# Global definitions
__application__ = _("StackApplet")
__authors__	    = ["Nathan Osman", "Isaiah Heyer"]
__artists__	    = ["Marco Ceppi"]
__translators__ = _("translator-credits")
__copyright__   = _("Copyright (c) 2011, Nathan Osman")
__license__	    = _('''Copyright (c) 2011 Nathan Osman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.''')
__version__	 = "1.5"

# Before we do much else, we are going to
# suppress any warnings - these wreak havok
# on Windows.
if platform.system() == 'Windows':
	import warnings
	warnings.simplefilter('ignore')

# modules required for using GTK
import pygtk
pygtk.require('2.0')
import gtk
import gobject

# Now for our own modules
# =======================

# First try to import the native app
# indicator module. Failing that, use our
# fallback module instead.
try:
	import appindicator
except ImportError:
	import appindicator_replacement as appindicator

# Do the exact same thing with pynotify.
# Use our fallback module if it's not available
try:
	import pynotify
except ImportError:
	import pynotify_replacement as pynotify

# ... the rest of them.
import config_store    # stores configuration settings / data
import defaults        # list of default settings
import network_thread  # for making anychronous network requests to the API
import preferences     # a web server that allows preferences to be changed

# Messaging menu support requires
try:
	import pymm
	CAN_USE_MM = 1
	
except ImportError:
	CAN_USE_MM = 0

#=======================================
# This is the root class that contains
# all of the core logic for StackApplet
#=======================================
class stackapplet:
	
	# Our constructor
	def __init__ (self):
		
		# Begin by creating the application indicator which we
		# use for displaying the icon.
		self.indicator = appindicator.Indicator ("stackapplet-desktop-client",
		                                         "stackapplet_grey",
		                                         appindicator.CATEGORY_APPLICATION_STATUS)
		
		self.indicator.set_status(appindicator.STATUS_ACTIVE)
		self.indicator.set_attention_icon("stackapplet")
		
		# Create the settings object so that we
		# can retrieve the list of accounts, preferences,
		# and so on.
		self.prefs = config_store.config_store()
		self.prefs.load_settings()
		
		self.refresh_rate = self.prefs.get_val("refresh_rate", defaults.__default_refresh_rate__)
		
		# Determine which icon set to use.
		if self.prefs.get_val("theme", defaults.__default_theme__) == 'dark':
			icon_to_use = 'stackapplet_grey'
		else:
			icon_to_use = 'stackapplet_light'
		
		self.indicator.set_icon(icon_to_use)
		
		# Create the menus for the accounts
		self.initialize_menu()
		
		# Create the messaging menu entry as well.
		if CAN_USE_MM:
			self.mm = pymm.pymm("/usr/share/applications/stackapplet.desktop")
		
		# Create the threaded network manager
		self.network = network_thread.network_thread()
		
		# Register one of our member functions with
		# the preference server so that we can supply
		# the requested details as necessary.
		preferences.register_callback(self.preference_callback)
		
		# Finally, begin updating the accounts.
		self.timer = None
		self.update()
	
	def initialize_menu(self):
		
		# Create the root GTK menu
		self.root_menu = gtk.Menu()
		
		# Create the Accounts and Tools
		# menu headers
		menu_item_accounts_header = gtk.MenuItem(_("Accounts"))
		menu_item_accounts_header.set_sensitive(False)
		menu_item_accounts_header.show()
		self.root_menu.append(menu_item_accounts_header)
		
		# Now add the accounts to the list.
		# The return value tells us if any accounts
		# were added, and if so, we add a separator
		if self.populate_menu():
			separator = gtk.SeparatorMenuItem()
			separator.show()
			self.root_menu.append(separator)
			
		else:
			menu_item_accounts_header.set_label(_("No Accounts"))
		
		# Now we add the remaining menu items
		menu_item_tools_header = gtk.MenuItem(_("Tools"))
		menu_item_tools_header.set_sensitive(False)
		menu_item_tools_header.show()
		self.root_menu.append(menu_item_tools_header)
		
		# For these items, we can use the stock
		# GTK menuitems
		menu_item_refresh = gtk.ImageMenuItem(stock_id=gtk.STOCK_REFRESH)
		menu_item_prefs   = gtk.ImageMenuItem(stock_id=gtk.STOCK_PREFERENCES)
		menu_item_about   = gtk.ImageMenuItem(stock_id=gtk.STOCK_ABOUT)
		menu_item_quit    = gtk.ImageMenuItem(stock_id=gtk.STOCK_QUIT)
		
		menu_item_refresh.connect("activate", self.menu_refresh_activate)
		menu_item_prefs  .connect("activate", self.menu_prefs_activate)
		menu_item_about  .connect("activate", self.menu_about_activate)
		menu_item_quit   .connect("activate", self.menu_quit_activate)
		
		menu_item_refresh.show()
		menu_item_prefs  .show()
		menu_item_about  .show()
		menu_item_quit   .show()
		
		# Add them all to the menu
		self.root_menu.append(menu_item_refresh)
		self.root_menu.append(menu_item_prefs)
		self.root_menu.append(menu_item_about)
		self.root_menu.append(menu_item_quit)
		
		self.indicator.set_menu(self.root_menu)
	
	def populate_menu(self):
		
		# Reset our accounts container
		self.accounts = []
		
		# Loop through each of the accounts in the list
		entries = self.prefs.get_val("accounts", defaults.__default_accounts__)
		accounts_processed = 0
		
		for entry in entries:
			
			self.create_menu_item(entry)
			accounts_processed += 1
		
		return accounts_processed
	
	def create_menu_item(self, entry):
		
		# Create the menu item
		menu_entry = gtk.MenuItem()
		menu_entry.set_label(_("Loading..."))
		
		# Add it to the menu
		menu_entry.show()
		self.root_menu.insert(menu_entry, 1 + len(self.accounts))
		
		# Create the messaging menu entry
		if CAN_USE_MM:
			mm_source = pymm.pymm_source(entry["site_name"])
			mm_source.set_count(0)
		else:
			mm_source = None
		
		# Now we need to check and see what values
		# are present in the configuration file.
		reputation_on_last_poll = entry['reputation_on_last_poll'] if 'reputation_on_last_poll' in entry else 0
		last_comment_id         = entry['last_comment_id']         if 'last_comment_id'         in entry else 0
		last_timestamp          = entry['last_timestamp']          if 'last_timestamp'          in entry else 0
		unread_reputation       = entry['unread_reputation']       if 'unread_reputation'       in entry else 0
		unread_comments         = entry['unread_comments']         if 'unread_comments'         in entry else 0
		unread_answers          = entry['unread_answers']          if 'unread_answers'          in entry else 0
		notifications           = entry['notifications']           if 'notifications'           in entry else 0
		
		account = { "menu"      : menu_entry,           # the Gtk menu
		            "mm_source" : mm_source,            # the Messaging menu source
		            "site"      : entry['site'],        # the URL of the SE site with this account
		            "site_name" : entry['site_name'],   # the name of the SE site
		            "user_id"   : entry['user_id'],     # the user ID for the site
		            
		            "reputation_on_last_poll"   : reputation_on_last_poll, # the API will never return 0, so this is safe
		            "last_comment_id"           : last_comment_id,         # the ID of the last comment, again - never 0
		            "last_timestamp"	        : last_timestamp,          # the last timestamp for answers to questions
		            "unread_reputation"         : unread_reputation,       # the amount of rep. earned that hasn't been seen
		            "unread_comments"           : unread_comments,         # the num. of comments that haven't been seen
		            "unread_answers"	        : unread_answers,          # the num. of answers that haven't been seen
		            "notifications"             : notifications }          # whether this item has notifications or not
		
		# If there are any notifications, display
		# the indicator.
		if notifications:
			self.indicator.set_status(appindicator.STATUS_ATTENTION)
		
		# Set the handlers
		menu_entry.connect("activate", self.menu_item_activate, account)
		
		if CAN_USE_MM:
			mm_source.source.connect("user-display", self.messaging_item_activate, account)
		
		# Add this item to our account list
		self.accounts.append(account)
	
	def display_notification(self, title, message, image):
		
		n = pynotify.Notification(title, message, image)
		n.show()
	
	# Check to see if there are any pending notifications
	# for any of the accounts.
	def test_notifications(self):
		
		any_notifications = 0
		
		for account in self.accounts:
			
			if account['notifications']:
				
				any_notifications = 1
		
		if any_notifications:
			self.indicator.set_status(appindicator.STATUS_ATTENTION)
		else:
			self.indicator.set_status(appindicator.STATUS_ACTIVE)
	
	# This function is responsible for dispatching
	# the API requests and making sure that when the
	# requests are completed that they are processed
	# by the right functions.
	def update(self):
		
		# Remove the current timer
		if(not self.timer == None):
			gobject.source_remove(self.timer)
		
		for account in self.accounts:
			
			self.update_account(account)
		
		# Now run the update again after the specified
		# timeout has been reached
		self.timer = gobject.timeout_add(self.refresh_rate, self.update)
	
	def update_account(self, account):
		
		# Start by dispatching a request to get the
		# current reputation for this account.
		self.network.issue_request(account['site'],
		                           '/users/' + account['user_id'],
		                           self.update_reputation,
		                           account)
		
		self.network.issue_request(account['site'],
		                           '/users/' + account['user_id'] + '/mentioned',
		                           self.update_comments,
		                           account)
		
		self.network.issue_request(account['site'],
		                           '/users/' + account['user_id'] + '/questions',
		                           self.update_answers_1,
		                           account)
	
	def update_reputation(self, data, account):
		
		# First check to see if there is any new reputation
		current_rep = data['users'][0]['reputation']
		
		# Check to see if there has been a change
		if not current_rep == account['reputation_on_last_poll'] and not account['reputation_on_last_poll'] == 0:
			
			# Calculate the change
			change = current_rep - account['reputation_on_last_poll']
			
			# Add this new reputation to the amount of reputation
			# that the user has not seen yet.
			account['unread_reputation'] += change
			
			#TODO: translations
			# Display a notification
			if change > 0:
				message = "You have gained " + str(change) + " reputation."
			else:
				message = "You have lost " + str(abs(change)) + " reputation."
			
			self.display_notification(account["site_name"] + ":", message, 'firefox')
			
			# Last but not least, we need to set that this item
			# has a pending unread notification
			account['notifications'] = 1
			
			self.indicator.set_status(appindicator.STATUS_ATTENTION)
			
		# Update the menu captions
		self.set_menu_label(account, current_rep)
			
		# Set this as the new reputation.
		account['reputation_on_last_poll'] = current_rep
	
	def update_comments(self, data, account):
		
		if len(data['comments']):
			
			# If either we have checked for a comment before
			# and it matches the last one...
			if (not account["last_comment_id"] == 0) and (not account["last_comment_id"] == data['comments'][0]["comment_id"]):
				
				# Display a notification
				message = data['comments'][0]["owner"]["display_name"] + " has posted a comment to you."
				self.display_notification(account["site_name"] + ":", message, 'firefox')
				
				account['unread_comments'] += 1
				account['notifications'] = 1
				
				self.set_menu_label(account, account['reputation_on_last_poll'])
				
				self.indicator.set_status(appindicator.STATUS_ATTENTION)
			
			# We do have a new comment.
			account["last_comment_id"] = data['comments'][0]["comment_id"]
	
	def update_answers_1(self, data, account):
		
		# We are only half of the way there. We have the users'
		# questions... now get the answers to those questions.
		
		# Create a vectorized list of question ids
		id_array = []
		
		for question in data['questions']:
			id_array.append(str(question["question_id"]))
		
		if len(id_array) and account["last_timestamp"]:
			
			# Join the IDs
			id_str = ";".join(id_array)
			
			# Issue the second part of the request
			self.network.issue_request(account['site'],
				                       '/questions/' + id_str + '/answers',
				                       self.update_answers_2,
				                       account,
				                       'fromdate=' + str(account["last_timestamp"]))
		
		# Set the timestamp
		account["last_timestamp"] = int(time.time())
	
	def update_answers_2(self, data, account):
		
		# See if there are any new answers
		if len(data['answers']):
			
			# Display a notification
			message = "Your have received " + len(data['answers']) + " new answer(s) to your questions."
			self.display_notification(account["site_name"] + ":", message, 'firefox')
			
			account['unread_answers'] += len(data['answers'])
			account['notifications'] = 1
			
			self.set_menu_label(account, account['reputation_on_last_poll'])
			
			self.indicator.set_status(appindicator.STATUS_ATTENTION)
	
	def set_menu_label(self, account, reputation):
		
		# Construct the label we will be using for the account in the menu
		label = account['site_name'] + ' / ' + str(reputation)
		
		# TODO: fix these for translations
		# Check for unread reputation
		if account['unread_reputation']:
			label += ' [' + str(account['unread_reputation']) + ' reputation]'
		
		# Check for comments
		if account['unread_comments']:
			label += ' [' + str(account['unread_comments']) + ' comments]'
		
		# Check for new answers
		if account['unread_answers']:
			label += ' [' + str(account['unread_answers']) + ' answers]'
		
		if account['notifications']:
			label = "* " + label
		
		account['menu'].set_label(label)
		
		# Don't forget about the MM
		account['mm_source'].set_count(account['unread_comments'] + account['unread_answers'])
	
	def preference_callback(self, command, event, data):
		
		# Process the request that was made
		if command == 'get_settings':
		
			data['accounts']     = self.accounts[:]
			data['theme']        = str(self.prefs.get_val('theme', defaults.__default_theme__))
			data['refresh_rate'] = int(self.refresh_rate)
			
		elif command == 'add_account':
			
			# Add the account
			self.create_menu_item(data)
			
			self.update_account(self.accounts[len(self.accounts) - 1])
			
		elif command == 'remove_account':
			
			index = int(data)
			
			# Remove the menu item and MM entry
			self.root_menu.remove(self.accounts[index]['menu'])
			
			if CAN_USE_MM:
				del self.accounts[index]['mm_source']
			
			# Delete the account
			del self.accounts[index]
			
			self.test_notifications()
		
		elif command == 'reorder':
			
			# We need to grab a copy of the item
			account = self.accounts[int(data['old_index'])]
			
			# Now remove it from the list
			self.accounts.remove(account)
			
			# Insert it into its new location
			self.accounts.insert(int(data['new_index']), account)
			
			# We also need to switch the menu items around too.
			self.root_menu.remove(account['menu'])
			self.root_menu.insert(account['menu'], 1 + int(data['new_index']))
		
		elif command == 'set_theme':
			
			if data == 'dark':
				icon_to_use = 'stackapplet_grey'
			else:
				icon_to_use = 'stackapplet_light'
			
			self.indicator.set_icon(icon_to_use)
			self.prefs.set_val('theme', str(data))
		
			self.indicator.set_icon(icon_to_use)
		
		# Set the event to allow the other thread
		# to continue along.
		event.set()
	
	def menu_item_activate(self, widget, account):
		
		# We need to either reset any unread stuff
		# or if there isn't any, then launch the user's
		# profile page.
		if account['notifications']:
			
			# Reset everything.
			account['notifications']     = 0
			account['unread_reputation'] = 0
			account['unread_comments']   = 0
			account['unread_answers']    = 0
			
			# Reset the menu caption
			self.set_menu_label(account, account['reputation_on_last_poll'])
			
			# Recheck for global notifications
			self.test_notifications()
		
		# Otherwise, display their profile page
		else:
			
			webbrowser.open("http://" + account["site"] + ".com/users/" + str(account["user_id"]))
	
	# Note that I have no clue what the third parameter is since
	# it is not documented ANYWHERE.
	def messaging_item_activate(self, indicator, unknown_data, account):
		
		# Just turn around and call menu_item_activate
		self.menu_item_activate(None, account)
	
	def menu_refresh_activate(self, widget):
		
		self.update()
	
	def menu_prefs_activate(self, widget):
		
		webbrowser.open("http://localhost:8140/")
	
	def menu_about_activate(self, widget):
		
		global __application__,__author__,__copyright__,__license__,__version__
		
		# Display the about dialog box
		dialog = gtk.AboutDialog()
		
		dialog.set_name(__application__)
		dialog.set_authors(__authors__)
		dialog.set_artists(__artists__)
		dialog.set_translator_credits(__translators__)
		dialog.set_copyright(__copyright__)
		dialog.set_license(__license__)
		dialog.set_version(__version__)
		
		if platform.system() == 'Windows':
			dialog.set_logo(gtk.gdk.pixbuf_new_from_file(os.path.join(sys.path[0], "../images/stackapplet.png")))
		else:
			dialog.set_logo_icon_name("stackapplet")
		
		dialog.connect("response",lambda x,y: dialog.destroy())
		
		dialog.run()
	
	def menu_quit_activate(self, widget):
		
		self.shutdown()
	
	# This is the safe shutdown function that
	# ensures all accounts' notification data
	# are saved with the accounts' information.
	def shutdown(self):
		
		# Save the accounts
		accounts_data = []
		
		for account in self.accounts:
			
			del account['menu']
			del account['mm_source']
			
			accounts_data.append(account)
		
		self.prefs.set_val('accounts', accounts_data)
		self.prefs.save_settings()
		
		gtk.main_quit()

stackapplet_instance = None

# SIGTERM handler
def signal_handler(a, b):
	
	global stackapplet_instance
	stackapplet_instance.shutdown()

# Create the instance of stackapplet and
# run the main GTK thread
if __name__ == "__main__":

	stackapplet_instance = stackapplet()
	
	signal.signal(signal.SIGTERM, signal_handler)
		
	gtk.main()
